# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sha
from urllib import urlencode
from xml.sax.saxutils import XMLGenerator
from xml.sax.handler import ContentHandler
from xml.sax import make_parser
from cStringIO import StringIO
import os
import url
"""
yarn.auth
This module provides classes for authentication.
"""

class Credentials(dict):
    fields = []

    def __init__(self, *args):
        if len(args) > len(self.fields):
            raise TypeError, "Too many arguments (got %i, expected %i.)" % (
                len(args), len(self.fields))
        else:
            dict.__init__(self)
            for field in self.fields:
                self[field] = ''
            for i in range(len(args)):
                self[self.fields[i]] = args[i]
                
    def __setitem__(self, key, val):
        if not key in self.fields:
            raise KeyError, ("'%s' is not a valid key (%s)." % (
                key, self.fields))
        # setting an empty value deletes the key
        if val:
            dict.__setitem__(self, key, val)
        else:
            if self.has_key(key):
                del(self[key])
                
    def getDigest(self):
        """
        return a hash identifying the contents of these Credentials,
        or an empty string if these Credentials contain no infomation.
        """
        argString = urlencode(self)
        if argString:
            return sha.new(argString).hexdigest()
        else:
            return ""

class AnonymousCredentials(Credentials):
    pass
        
class UserPassCredentials(Credentials):
    fields = ['username', 'password']

class TokenCredentials(Credentials):
    fields = ['token']

# get list of all classes implementing Credentials
import types
credTypes = {}
for name, item in locals().items():
    if type(item) in (types.TypeType, types.ClassType):
        if issubclass(item, Credentials):
            credTypes[name] = item

class Keychain(dict):
    """
    Keychain object - keeps track of which credentials to use
    for specific urls, as a dict of url:creds.
    """
    def __init__(self, filename=None):
        dict.__init__(self)
        self.filename = filename
        if self.filename and os.path.exists(self.filename):
            self.loadXml(file(self.filename))

    def __setitem__(self, key, val):
        # convert Urls to strings
        if isinstance(key, url.Url):
            key = str(key)
        else:
            key = str(url.Url(key))
        return dict.__setitem__(self, key, val)

    def __getitem__(self, key):
        return dict.__getitem__(self, str(key))

    def get(self, key, default=None):
        # try to find credentials that match part of the key
        searchUrl = url.Url(key)
        bestMatch = None
        for u in self.keys():
            matchUrl = url.Url(u)
            if matchUrl.protocol == searchUrl.protocol and matchUrl.server == searchUrl.server and matchUrl.port == searchUrl.port and searchUrl.path.startswith(matchUrl.path):
                # keep the url with the most-specific path.
                if (not bestMatch) or len(matchUrl.path) > len(bestMatch.path):
                    bestMatch = matchUrl

        if bestMatch:
            return self[bestMatch]
        else:
            return default
    
    def getXml(self):
        # use SaxWriter stuff
        stream = StringIO()
        xml = XMLGenerator(stream, 'utf-8')
        xml.startElement('keychain', {})
        for url, creds in self.items():
            attrs = dict(creds.items())
            attrs['url'] = url
            attrs['type'] = creds.__class__.__name__
            xml.startElement('credentials', attrs)
            xml.endElement('credentials')
        xml.endElement('keychain')
        stream.seek(0)
        return stream.read()

    def loadXml(self, stream):
        parser = make_parser()
        parser.setContentHandler(_KeychainXmlParser(self))
        parser.parse(stream)

    def save(self):
        if not self.filename: raise Exception, "No filename to save to!"
        xml = self.getXml()
        file(self.filename, 'w+b').write(xml+'\n')


class _KeychainXmlParser(ContentHandler):
    def __init__(self, keychain):
        self.keychain = keychain
        ContentHandler.__init__(self)

    def startElement(self, name, attrs):
        if name == 'credentials':
            u = ''
            fields = {}
            for attrName, val in attrs.items():
                type = 'UserPassCredentials'
                if attrName == 'url':
                    u = val
                elif attrName == 'type':
                    type = val
                else:
                    fields[attrName] = val
            # select the correct Cred class.
            cred = credTypes[type]()
            for field, val in fields.items():
                cred[field] = val
                    
            self.keychain[url.Url(u)] = cred 
