# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re
from xml.dom import minidom
from yarn import proxy, formats, url
import os.path
from urlparse import urljoin

class OPMLParser(formats.Parser):
    priority = 10
    features = ["createpath", "deletepath", "createlink", "deletelink"]
    
    def canParse(stream, metadata):
        stream.seek(0)
        validMimeTypes = [
            'text/x-opml', 
            'application/xml', 
            'text/xml', 
            'text/html', 
            'text/plain',
            '']
        if metadata.get('mimetype') == 'text/x-opml':
            return 1
        elif metadata.get('filename', '').endswith('.opml'):
            return 1
        elif metadata.get('mimetype', '') not in validMimeTypes:
            return 0
        elif re.match("<\?xml.*<opml",
                    stream.read(1000), re.IGNORECASE+re.DOTALL):
            return 1
        else:
            return 0
    canParse = staticmethod(canParse)
    
    def __init__(self, stream, metadata):
        self.pathMap = {}
        self.metadata = metadata
        stream.seek(0)
        self.document = minidom.parse(stream)
        self.filename = metadata.get('filename')
        self.url = metadata.get('url', '')
        self.writeable = False
        if self.filename and os.path.isfile(self.filename):
            self.writeable = True

    def listMessages(self, path):
        return []
    
    def getMessage(self, path, key):
        raise KeyError, "No such message"
        
    def getName(self, path):
        if path == '':
            headerEl = self.document.getElementsByTagName('head')[0]
            titleEls = headerEl.getElementsByTagName('title')
            if titleEls:
                return titleEls[0].firstChild.data
            else:
                return self.metadata['url']
        else:
            node = self._getPathNode(path)
            return node.getAttribute('text') or node.getAttribute('title', '')
        
    def _getPathNode(self, path):
        parent = self.document.getElementsByTagName('body')[0]
        pathParts = path and path.split('/') or []
        while pathParts:
            nextPart = pathParts.pop(0)
            newParent = None
            for el in parent.childNodes:
                if el.nodeType != 1: continue # ignore text/comment nodes
                if el.getAttribute('text') == nextPart:
                    newParent = el
            if not newParent: raise "InvalidPathError"
            parent = newParent
        return parent

    def _listChildren(self, path):
        "return list of name, url for elements in the given path."
        pathNode = self._getPathNode(path)
        for el in pathNode.childNodes:
            if el.nodeType != 1: continue # skip text and comment nodes.
            url = el.getAttribute('xmlUrl') or el.getAttribute('url') or ''
            name = el.getAttribute('text') or el.getAttribute('title') or ''
            yield name, url

    def listPaths(self, path):
        paths = []
        for name, url in self._listChildren(path):
            if not url:
                paths.append(name)
        return paths

    def getLinks(self, path):
        links = {}
        for name, u in self._listChildren(path):
            if u:
                # resolve relative links
                fullUrl = url.Url(self.metadata['url']).getUrl(u)
                links[u] = url.Link(name, fullUrl, 'child')
        return links.values()
        
    def getUrl(self, path):
        return self.url + '#' + path
        
    def createLink(self, parentPath, link):
        if not self.writeable: raise "ReadOnlyError"
        try:
            # raise an error if the link already exists 
            self._getItemAndParent(parentPath, link)
            raise Exception, "link already exists!"
        except KeyError:
            pass
        el = self.document.createElement('outline')
        el.setAttribute('text', link.title)
        el.setAttribute('xmlUrl', str(link.url))
        self._getPathNode(parentPath).appendChild(el)
        self.save()
        return link

    def _getItemAndParent(self, parentPath, link=None, folder=None):
        if not link and not folder: raise Exception, "need link or folder"
        parentNode = self._getPathNode(parentPath)
        for node in parentNode.childNodes:
            if node.nodeType != 1: continue
            url = node.getAttribute('xmlUrl') or node.getAttribute('url')
            name = node.getAttribute('text') or node.getAttribute('title')
            if (link and url == link.url and name == link.title) or (
                folder and name == folder and not url):
                    return parentNode, node
        raise KeyError, link.title

    def deleteLink(self, path, link):
        parent, child = self._getItemAndParent(path, link=link)
        parent.removeChild(child)
        self.save()

    def editLink(self, parentPath, oldLink, newLink):
        parent, node = self._getItemAndParent(parentPath, link=oldLink)
        for attr in 'text', 'title':
            if node.getAttribute(attr):
                node.setAttribute(attr, newLink.title)
        for attr in 'xmlUrl', 'url':
            if node.getAttribute(attr):
                node.setAttribute(attr, str(newLink.url))
        self.save()
        return newLink

    def deleteFolder(self, parentPath, folder):
        parent, child = self._getItemAndParent(parentPath, folder=folder)
        parent.removeChild(child)
        self.save()

    def save(self):
        if not self.writeable: raise Exception, "OPML File is read-only"
        xml = self.document.toxml(encoding=self.document.encoding or 'utf-8')
        file(self.filename, 'w+b').write(xml)
