# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import time
import email
import email.Utils, email.Encoders
import email.Message
from email.Header import Header, decode_header
from email.MIMEText import MIMEText
from email.MIMEBase import MIMEBase
from email.MIMENonMultipart import MIMENonMultipart
import re
import md5
import sys
from extlib.textile import textile
from extlib import stripogram
from url import Url
import yarn 
import types

def html2safehtml(html, tags=[]):
    try: output = stripogram.html2safehtml(html, tags)
    except: output = html
    return output
    
def html2text(html):
    try: text = stripogram.html2text(html)
    except:
        # fall back to simple-and-hacky conversion
        p = re.compile(r"<p.*?>", re.IGNORECASE | re.MULTILINE | re.DOTALL)
        br = re.compile(r"<br.*?>", re.IGNORECASE | re.MULTILINE | re.DOTALL)
        alltags = re.compile(r"<.*?>",
                             re.IGNORECASE | re.MULTILINE | re.DOTALL)
        text = html.replace("\r","")
        text = re.sub(p,"\n\n", text)
        text = re.sub(br,"\n", text)
        text = re.sub(alltags, "", text)
    return text

class NameAndAddress(object):
    def __init__(self, addrStr=""):
        self.name, self.address = email.Utils.parseaddr(addrStr)
            
    def __unicode__(self):
        return email.Utils.formataddr((self.name, self.address))

    def __str__(self):
        # this will raise a UnicodeEncodeError if name or addr is non-ascii.
        return str(self.__unicode__())

    def encode(self, *args): return unicode(self).encode(*args)

    def __iter__(self): return iter(unicode(self))
    
        
class AlternativeContentSet(object):
    def __init__(self, mimeMessage):
        self.mimeMessage = mimeMessage

    def _getAlternativeParts(self):
        parent = self.mimeMessage
        while parent.is_multipart():
            if parent.get_content_type() == 'multipart/alternative':
                return parent.get_payload()
            else:
                parent = parent.get_payload()[0]
        return [parent]

    def _getMultipartPayload(self):
        if self.mimeMessage.is_multipart():
            if self.mimeMessage.get_type() == 'multipart/alternative':
                return self.mimeMessage.get_payload()
            else:
                parts = self.mimeMessage.get_payload()
                if parts[0].is_multipart():
                    if parts[0].get_type() == 'multipart/alternative':
                        return parts[0].get_payload()
                    
    def isMultipart(self):
        return self._getMultipartPayload() and True or False

    def __len__(self):
        return len(self._getAlternativeParts())

    def __getitem__(self, itemNo):
        part = self._getAlternativeParts()[itemNo]
        return MessageContents(part)

    def __setitem__(self, itemNo, contents):
        newMessage = contents.mimeMessage
        if not self.isMultipart():
            if itemNo not in (0, -1):
                raise IndexError, "Index out of range"
            else:
                self.mimeMessage.set_payload(newMessage.get_payload(),
                                             newMessage.get_charset())
                self.mimeMessage.set_type(newMessage.get_type())
        else:
            self.getMultipartPayload[itemNo] = contents

    def __delitem__(self, items):
        for itemNo in range(*items.indices(len(self))):
            if itemNo not in (0, -1):
                raise ValueError, "Can't delete main contents!"
            elif not self.isMultipart():
                raise KeyError, itemNo
            else:
                del(self._getMultipartPayload()[itemNo])

    def _getPartsForInsert(self):
        target = self._getMultipartPayload()
        if not target:
            # no existing multipart section found, transform self.mimeMessage
            existingPart = MIMEBase(*self.mimeMessage.get_type().split('/'))
            existingPart.set_payload(self.mimeMessage.get_payload(),
                                     self.mimeMessage.get_charset())
            self.mimeMessage.set_type('multipart/alternative')
            self.mimeMessage.set_payload([existingPart])
            target = self.mimeMessage.get_payload()
        return target
            
    def append(self, contents):
        self._getPartsForInsert().append(contents.mimeMessage)

    def insert(self, index, contents):
        self._getPartsForInsert().insert(index, contents.mimeMessage)
        

class MessageContents(object):
    def __init__(self, contents=None, mimetype='text/plain'):
        if not contents:
            self.mimeMessage = email.MIMEBase.MIMEBase(*mimetype.split('/'))
        elif isinstance(contents, email.Message.Message):
            self.mimeMessage = contents
        elif hasattr(contents, 'read'):
            self.mimeMessage = email.message_from_file(contents)
        else:
            self.mimeMessage = email.message_from_string(str(contents))
    
    def _getMainPart(self):
        return self.alternatives[-1].mimeMessage

    def _getContents(self):
        part = self._getMainPart()
        charset='ascii'
        try:
            charset = str(part.get_content_charset() or 'ascii')
            contents = (part.get_payload(decode=1) or '').decode(charset)
        except TypeError, e:
            contents = ''
        if not contents:
            # this is a silly workaround for an email.Message problem.
            # if a Message is created as utf-8, it gets a
            # Content-Transfer-Encoding: base64 header, but the
            # actualy payload is not encoded... so get_payload(decode)
            # returns gibberish that can't be unencoded.  I'd like
            # to find a better solution for this, but no luck so far.
            print "MessageContents class says: %s is not my friend." % charset
            contents = (part.get_payload() or '').decode(charset)
        return contents

    def _smartEncode(self, contents, charset=None):
        """
        intelligently figure out a charset to use for encoding
        the given unicode string, and return the encoded
        string and charset.
        """
        if not charset:
            maxChar = contents and max([ord(c) for c in contents]) or 0
            if maxChar < 128:
                charset = 'us-ascii'
            elif maxChar < 256:
                charset = 'latin-1'
            else:
                charset = self._getMainPart().get_content_charset()
        if not charset: charset = 'utf-8'
        # at this point we're still not guaranteed to have an acceptible
        # charset... the user may have passed a unicode string unsupported
        # my the message's current charset.
        try:
            return contents.encode(str(charset)), charset
        except UnicodeEncodeError, e: # fallback to utf-8
            return contents.encode('utf-8'), 'utf-8'

    def _setContents(self, contents):
        contents, charset = self._smartEncode(contents)
        # this removes existing alternative parts
        del(self.alternatives[:-1])
        self._getMainPart().set_payload(contents, charset)

    contents = property(_getContents, _setContents)
    mimetype = property(lambda self: self._getMainPart().get_type(),
                        lambda self, type: self._getMainPart().set_type(type))
    alternatives = property(
        lambda self: AlternativeContentSet(self.mimeMessage))


class Message(MessageContents):

    def __init__(self, *args, **kwargs):
        MessageContents.__init__(self, *args, **kwargs)
        # set timestamp to now
        self.setTimestamp(time.localtime(time.time()))
            
    def __getitem__(self, key):
        value = self.mimeMessage.get(key)
        if value:
            decoded = decode_header(value)
            if decoded: 
                value, encoding = decoded[0]
                value = value.decode(str(encoding or 'ascii'))
        return value or ''

    def get(self, key, default=''):
        return self[key] or default
        
    def __setitem__(self, key, value):
        key = key.lower()
        headerName = key
        value, charset = self._smartEncode(value)
        if charset == 'us-ascii':
            value = str(value)
        else:
            value = Header(value, charset)

        del(self.mimeMessage[headerName])
        self.mimeMessage[headerName] = str(value)

    def has_key(self, key):
        result = self.mimeMessage.has_key(key)
        return result

    def getAllHeaders(self):
        headers = {}
        for key in self.keys():
            headers[key] = self.get(key, '')
        return headers
        
    def getTimestamp(self):
        return yarn.parseDate(self.get('Date', ''))
        
    def setTimestamp(self, value):
        timestamp = value
        # detect the format of the timestamp and convert it
        # to email format.
        # TODO: support formats like 2003-03-29T14:40:38-08:00
        ts = None
        if not timestamp:
            pass
        elif type(timestamp) in (type([]), type((0,)), time.struct_time):
            if len(timestamp) == 9:
                    # localtime 9-tuple format
                ts = email.Utils.formatdate(time.mktime(timestamp))
        elif re.match('\d{8}T\d\d:\d\d:\d\d', timestamp):
            # xmlrpc format
            dateStr, timeStr = timestamp.split('T')
            yr, mo, day = dateStr[0:4], dateStr[4:6], dateStr[6:8]
            hr, min, sec = timeStr[0:8].split(":")
            timebits = [yr, mo, day, hr, min, sec, 0, 0, 0]
            # convert all parts to ints
            for bit in timebits:
                timebits[timebits.index(bit)] = int(bit)          
            ts = email.Utils.formatdate(time.mktime(timebits))
        else:
            parsed = email.Utils.parsedate(timestamp)
            if parsed: ts = timestamp
        if ts:
            self['Date'] = ts
        else:
            raise "Bad date format: %s" % timestamp
            
    def getMessageId(self):
        """
        return a Message-ID, generating one based on the message contents
        if there's is no Message-ID field in the message's headers.
        """
        if not self.get('message-id', ''):
            # generate an ID based on the hash of this message's
            # from: to: subject: headers + the body.
            hash = md5.new()
            for header in ['from', 'to', 'subject']:
                hash.update(self.get(header, '').encode('utf-8', 'ignore'))
            hash.update(self.contents.encode('utf-8', 'ignore'))
            self.mimeMessage['Message-Id'] =  "<%s.md5@localhost>" % (
                hash.hexdigest())
        return str(self.mimeMessage['Message-Id'])

    def _createPart(self, contents, mimetype, charset):
        mainType, subType = mimetype.split('/')
        if mainType == 'text':
            newPart = email.MIMEText.MIMEText(contents, subType, charset)
        else:
            newPart = MIMENonMultipart(mainType, subType)
            newPart.set_charset(charset)
            newPart.set_payload(contents)
        return newPart

    def getCopy(self, generateMissingParts=0):
        "get an email.Message.Message containing the contents of this message."
        
        from copy import deepcopy
        m = deepcopy(self) # make a copy so we don't mess up self

        if generateMissingParts:
            if m.mimetype == 'text/plain': # FIXME: or whatever/rfc822
                if m['X-Yarn-Link']:
                    m.contents += '\n\n%s\n' % m['X-Yarn-Link']
            elif self.mimetype == 'text/html':
                if m['X-Yarn-Link']: #and not html.count(self['link']):
                    endOfBody = re.compile('(</(body|html)>.*|\\Z)',
                                 re.IGNORECASE | re.MULTILINE | re.DOTALL)
                    link = self['X-Yarn-Link'].encode('ascii', 'replace')
                    linkHTML = u'<p><a href="%s">%s</a></p>' % (link, link)
                    m.contents = endOfBody.sub(('%s\\1' % linkHTML),
                                               m.contents)

                # generate 'missing' text version.
                currentTypes = [a.mimetype for a in m.alternatives]
                if not 'text/plain' in currentTypes:
                    textContents = MessageContents()
                    textContents.contents = html2text(m.contents)
                    m.alternatives.insert(0, textContents)
        return m

    def asString(self, unixFrom=0, generateMissingParts=0):
        "get an rfc822 email-formatted version of this message."
        cp = self.getCopy(generateMissingParts)
        return cp.mimeMessage.as_string(unixFrom)
        
    def __str__(self):
        return self.asString()

    def __cmp__(self, other):
        return cmp(str(self), str(other))

