# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from yarn import formats, proxy, message, url, features, events
import os
import os.path
import re
from mailbox import Maildir
from socket import gethostname
import time
from twisted.internet import defer, reactor
from urllib import unquote

class FileProxySmartOpener(proxy.Proxy):
    protocols = ['file']

    def connect(self):
        self._realProxy = None
        filename = self.url.path
        if os.path.isdir(filename):
            if isMaildir(filename):
                self._realProxy = MaildirProxy(self.url, self.connector)
            else:
                self._realProxy = DirectoryStore(url, connector)
        else:
            self._realProxy = SingleFileProxy(self.url, self.connector)
        return self._realProxy.connect()

    def __getattribute__(self, name):
        if name != 'connect':
            try:
                prox = object.__getattribute__(self, '_realProxy')
                return getattr(prox, name)
            except AttributeError, e:
                # _realProxy isn't set, or didn't have the attribute
                pass
        return object.__getattribute__(self, name)

class SingleFileProxy(proxy.SimpleProxy):
    
    def getFile(self):
        d = defer.Deferred()
        reactor.callLater(0, self._reallyGetFile, d)
        return d

    def _reallyGetFile(self, finished):
        try:
            results = (file(self.url.path), {'url': self.url,
                                             'filename': self.url.path})
            finished.callback(results)
        except Exception, e:
            finished.errback(e)

def getUserName(fileName):
    uid = os.stat(fileName).st_uid
    try:
        import pwd
        userName = pwd.getpwuid(uid)[4].split(',')[0]
    except:
        userName = uid
    return userName


class oldstuff:
    def blah():
        path = self.url.path
        f = open(path)
        pars
        er = formats.parseFile(f, {'filename': path})
        self.properties['name'] = parser.getHeaders().get('name') or self.url.path
        messages = list(parser.getMessages())
        f.close()
        if len(messages) == 1 and not messages[0].author:
            messages[0].author.name = getUserName(path)
        
        for n in range(len(messages)):
            self.messages[str(n)] = messages[n]
            
        for childInfo in parser.getChildUrls():
            path = childInfo['name']
            if childInfo.get('autoSubscribe'):
                self.recommendedPaths.append(path)
                del(childInfo['autoSubscribe'])
            self.paths[path] = childInfo
        
        return defer.succeed(1)
        
class DirectoryStore(proxy.Proxy):
    def load(self):
        path = self.url.path
        fileList = os.listdir(path)
        for fileName in fileList:
            self.name = fullName = os.path.join(path, fileName)
            if os.path.isfile(fullName):
                file = open(fullName)
                messages = formats.parseFile(file, {'name':fullName})
                file.close()
                for n in range(len(messages)):
                    message = self.messages[n]
                    if not message.author:
                        message.author.name = getUserName(fullName)
                    self.messages[n] = message
                return 1
    
def isMaildir(path):
    maildirDirs = 0
    for folder in ('cur', 'new', 'tmp'):
        if os.path.isdir(os.path.join(path, folder)):
            maildirDirs += 1
    if maildirDirs == 3:
        return 1
    else:
        return 0
            
class MaildirProxy(proxy.Proxy):
    
    def __init__(self, url, connector, credentials=None):
        proxy.Proxy.__init__(self, url, connector, credentials)
        if hasattr(os.path, 'realpath'):
            self.path = os.path.realpath(url.path)
        else:
            self.path = url.path
        parent, thisdir = os.path.split(self.path)
        if thisdir == 'mbox':
            self.name = os.path.split(parent)[1]
        else:
            self.name = unquote(thisdir)
        self.paths = {'': self.path} # dict of {'path': '/full/dir/path'}

    def connect(self):
        self._lookupPaths('') # load base path list
        return defer.succeed(None)

    def getFeatures(self, path):
        # FIXME: maybe check for read-only?
        return [
            features.MESSAGES,
            features.POST_MESSAGE,
            features.DELETE_MESSAGE,
            features.EDIT_MESSAGE,
            features.CREATE_FOLDER,
            features.DELETE_FOLDER,
            features.MOVE_FOLDER,
            ]

    def _parseMessageFile(self, fp):
        m = message.Message(fp)
        return m
        
    def listMessages(self, path):
        mbox = Maildir(self.paths[path], lambda _: None)
        keys = []
        self.messageCount = 0
        self.newMessageCount = 0
        while len(mbox.boxes) > 0:
            fileName = mbox.boxes[0]
            #if fileName.split(os.sep)[-2] == 'new':
            #    self.newMessageCount += 1
            keys.append(fileName.split(os.sep)[-1])
            mbox.next()
        return defer.succeed(keys)

    def _lookupPaths(self, path):
        parentDir, thisDir = os.path.split(self.path)
        for dir in os.listdir(self.path):
            fullPath = os.path.join(self.path, dir)
            if isMaildir(fullPath):
                if dir[0] == '.':
                    pathName = joinPath(path, dir[1:])
                else:
                    pathName = joinPath(path, dir)
                self.paths[pathName] = fullPath

    def listPaths(self, path):
        self._lookupPaths(path) # refresh path list
        paths = []
        for p in self.paths.keys():
            if p.startswith(path) and (p.count('/') == path.count('/') + 1):
                paths.append(p)
        return defer.succeed(paths)

    def _getFullFilename(self, path, fileName):
        "find the file by this name in the cur/ or new/ directories"
        curFile = os.path.join(self.paths[path], 'cur', fileName)
        newFile = os.path.join(self.paths[path], 'new', fileName)
        if os.path.exists(curFile):
            return curFile
        elif os.path.exists(newFile):
            return newFile
        else:
            raise KeyError, fileName
        
    def getMessage(self, path, fileName):
        fullFileName = self._getFullFilename(path, fileName)
        m = self._parseMessageFile(file(fullFileName))
        return defer.succeed(m)

    def _getNewMessageFileName(self, msg):
        attempt = 0
        while attempt < 5:
            fileName = ".".join([str(time.time()),
                                 str(os.getpid()),
                                 gethostname()])
            #if msg.isNew:
            fileName = os.path.join(self.path, "new", fileName)
            #else:
            #    fileName = os.path.join(self.path, "cur", fileName)
            if os.path.exists(fileName):
                time.sleep(.1)
                attempt = attempt + 1
            else:
                return fileName  
        raise Exception, "Unable to generate file name!"  

    def postMessage(self, path, msg):
        print "IMAP: posting", repr(msg)
        fileName = self._getNewMessageFileName(msg)
        f = file(fileName, "w+b")
        f.write(str(msg))
        f.close()
        key = fileName.split(os.sep)[-1]
        self.handleEvent(path, events.MESSAGE_POSTED, key)
        return defer.succeed(key)
    
    def edit(self, fileName, msg):
        oldFileName = self.getFullFilename(fileName)
        fileNameParts = oldFileName.split(os.sep)
        if not msg.isNew:
            fileNameParts[-2] = 'cur'
        else:
            fileNameParts[-2] = 'new'      
        f = open(newFileName, "w")
        f.write(str(msg))
        f.close()
        if newFileName != oldFileName:
            if msg.isNew:
                self.newMessageCount += 1
            else:
                self.newMessageCount -= 1
            # delete the old file.
            os.unlink(oldFileName)
            if oldFileName in self.files: 
                self.files.remove(oldFileName)
                # fixme: how could newFilename already be in self.files???
                self.files.append(newFileName)
        return defer.succeed(fileName)
        
    def markMessageAsRead(self, fileName):
        m = self.getMessageByID(fileName)
        m.isNew = 0
        return self.edit(fileName, m)
            
    def deleteAll(self):
        for messageDir in ['cur','new']:
            for fileName in os.listdir(os.path.join(self.path, messageDir)):
                filePath = os.path.join(self.path, messageDir, fileName)
                os.remove(filePath)
        # refresh messageCount, etc
        self.listMessages()
        return 1
    
    def delete(self, fileName):
        d = defer.Deferred()
        os.remove(self.getFullFilename(fileName))
        if fileName in self.files:
            self.files.remove(fileName)
        # refresh MessageCount, etc
        self.listMessages()
        return defer.succeed(fileName)
