# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.web import xmlrpc
from twisted.internet import defer
from twisted.python import failure
from yarn import message, proxy, errors, url, features, auth
import re, time

"""
support for the Blogger (and MetaWeblog) API
see http://plant.blogger.com/api/RPC2
"""

class MetaWeblogProxy(proxy.Proxy):
    protocols = ['metaweblog']
    
    appkey = '3696CDCBCADAE48B6CD3DD229EFF7A320F826B60'
    canPost = 1
    defaultRefreshFrequency = 1
    encoding = 'utf-8' # used by Python's xmlrpclib.  If the server
                       # doesn't also support utf-8 there will be problems.
    refreshTime = 10 # only cache results for this number of seconds
    
    def __init__(self, *args, **kwargs):
        proxy.Proxy.__init__(self, *args, **kwargs)
        self.blogs = {}
        self.authors = {}
        self.messageCache = {}
        self.timestamps = {}
        self.blog = self.url.fragment
        self.webUrl = url.Url(self.url.path)
        self.server = xmlrpc.Proxy(self.url.path)
        self.name = ""
        
    def connect(self):
        d1 = self.server.callRemote("blogger.getUsersBlogs", self.appkey,
                                    self.credentials.get('username', ''),
                                    self.credentials.get('password', ''),
                                    ).addCallback(self._gotBlogs)
        d1.addErrback(self._catchRpcError)
        d2 = self.server.callRemote("blogger.getUserInfo", self.appkey,
                                    self.credentials.get('username', ''),
                                    self.credentials.get('password', ''),
                                    ).addCallback(self._gotUserInfo)
        d2.addErrback(self._catchRpcError)
        allTasks = defer.DeferredList(
            [d1, d2]).addCallback(
            self._checkConnectResults).addCallback(self._gotServerInfo)
        # catch the errors once the deferredList is done with them...
        # otherwise they end up floating around in the ether and Twisted
        # thinks something went wrong.
        d1.addErrback(lambda failure: None)
        d2.addErrback(lambda failure: None)
        return allTasks

    def _checkConnectResults(self, results):
        for success, val in results:
            if not success:
                return self._catchRpcError(val)
        return 1

    def _gotServerInfo(self, results):
        author = self.authors.values()[0]
        if author.get('firstname') and author.get('lastname'):
            username = "%s %s" % (author['firstname'], author['lastname'])
        elif author.get('firstname'):
            username = author['firstname']
        else:
            username = self.credentials.get('username', '')
        self.name = "%s weblogs (MetaWeblog API)" % (self.webUrl.server)

    def _catchRpcError(self, e):
        # try to guess what fails and raise a meaningful exception
        #print "GOT RPC ERROR:", e
        message = e.getErrorMessage()
        if re.match(".*(login|username|password).*", message, re.IGNORECASE):
            if self.credentials:
                return failure.Failure(errors.LoginFailed(
                    auth.UserPassCredentials, message))
            else:
                return failure.Failure(errors.CredentialsRequired(
                    auth.UserPassCredentials, message))
        else:
            return e
        
    def _gotBlogs(self, blogs):
        for blog in blogs:
            self.blogs[blog['blogName']] = blog

    def _gotUserInfo(self, info):
        self.authors[info['userid']] = info

    def listPaths(self, path):
        # return the list of blogs, unless they're already looking at one.
        if path == "" and not self.blog:
            return defer.succeed(self.blogs.keys())
        else:
            return defer.succeed([])

    def _getMessages(self, path, count=20):
        if not self.blogs.has_key(path): raise KeyError, "No such path"
        cachedSecs = time.time() - self.timestamps.get(path, 10000)
        if self.messageCache.has_key(path) and cachedSecs <= self.refreshTime:
            return defer.succeed(self.messageCache[path])
        else:
            return self.server.callRemote("metaWeblog.getRecentPosts",
                                        self.blogs[path]['blogid'],
                                       self.credentials.get('username', ''),
                                       self.credentials.get('password', ''),
                                       count,
                                       ).addCallback(self._gotPosts, path)

    def _gotPosts(self, posts, blog):
        self.timestamps[blog] = time.time()
        for post in posts:
            self._rememberPost(post, blog)
        return self.messageCache[blog]
    
    def _rememberPost(self, post, blog):
        messages = self.messageCache.setdefault(blog, {})
        m = message.Message(mimetype='text/html')
        m['Subject'] = post.get('title', '') 
        m.contents = post['description']
        authorInfo = self.authors.get(post['userid'])
        if not authorInfo:
            authorInfo = {
                'firstname': 'Author',
                'lastname': '%s' % (post['userid']),
                }
        m['from'] = '%s %s <%s>' % (authorInfo.get('firstname',''),
                                  authorInfo.get('lastname',''),
                                  authorInfo.get('email', ''))
        m['Message-Id'] = "<%s@%s/%s#%s>" % (post['postid'], self.url.server,
                                             self.url.path,
                                             self.blogs[blog]['blogid'])
        m.setTimestamp(post['dateCreated'].value)
        messages[post['postid']] = m

    def listMessages(self, path):
        if not path and not self.blog:
            return defer.succeed([])
        else:
            if self.blog: path = self.blog
            return self._getMessages(path).addCallback(lambda ms: ms.keys())
        
    def getMessage(self, path, messageKey):
        if self.blog: path = self.blog
        return self._getMessages(path).addCallback(lambda ms: ms[messageKey])
    
    def postMessage(self, path, message):
        if self.blog: path = self.blog
        contents = message.contents
        if message.mimetype == 'text/html':
            # FIXME: make this more reliable - use a parser instead of re
            # also it would be nice to detect the text filter in use
            # and work with it...
            start = re.compile(r"^.*(<body[^>]*>)",
                               re.IGNORECASE | re.MULTILINE | re.DOTALL)
            end = re.compile(r"(</body.*)",
                             re.IGNORECASE | re.MULTILINE | re.DOTALL)
            contents = end.sub('', start.sub('', contents))
        post = {
            'description': contents.encode(self.encoding),
            'title': message['Subject'].encode(self.encoding),
        }
        d = defer.Deferred()
        username = self.credentials.get('username', '').encode(self.encoding)
        password = self.credentials.get('password', '').encode(self.encoding)
        posting = self.server.callRemote("metaWeblog.newPost",
                                         self.blogs[path]['blogid'],
                                         username,
                                         password,
                                         post,
                                         xmlrpc.Boolean(1))
        posting.addCallback(self._postedMessage, path, d)
        posting.addErrback(self._catchRpcError).addErrback(
            lambda err: d.callback(err))
        return d

    def _postedMessage(self, postID, path, finished):
        fetching = self.server.callRemote("metaWeblog.getPost",
                                          postID,
                                          self.credentials.get('username', ''),
                                          self.credentials.get('password', ''),
                                          )
        fetching.addCallback(
            self._fetchedPostedMessage, path, finished).addErrback(
            finished.errback)

    def _fetchedPostedMessage(self, post, path, finished):
        self._rememberPost(post, path)
        finished.callback(post['postid'])

    def getUrl(self, path):
        if path == "":
            return self.url
        else:
            return url.Url(self.url + "#" + path)

    def getName(self, path):
        if not path:
            return self.blog or self.name
        else:
            return path

    def getFeatures(self, path):
        if not path and not self.blog:
            return [features.FOLDERS]
        else:
            return [
                features.MESSAGES,
                features.POST_MESSAGE,
                features.DELETE_MESSAGE,
                features.EDIT_MESSAGE,
                ]
