# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from yarn import message, proxy
from yarn.formats.mail import Parser
from twisted.internet.protocol import ClientFactory
from pop3support import POP3Client
from twisted.internet import reactor, defer
from cStringIO import StringIO
import time

PROTOCOLS = ['pop3']
PORTS = {'pop3': 110}

def openProxy(url, connector):
    prox = POP3Proxy(url, connector)
    return defer.succeed(prox)
    #return prox.connect().addCallback(lambda _: prox)

class POP3Proxy(proxy.Proxy, ClientFactory):
    
    def __init__(self, url, connector):
        proxy.Proxy.__init__(self, url, connector)
        self.messageMap = {} # map of uidl->index
        self.client = None
        self.commands = []
        self.protocol = BasicPOP3
        #self.features = [messaging.DELETE_MESSAGE]
        self.busy = 0
        
    def __del__(self):
        if self.client: self.client.quit()
        
    def getProperties(self, path=None):
        return defer.succeed({'name':
                "%s@%s POP" % (self.url.username, self.url.server)})
        
    def connectionFailed(self, error):
        for commandInfo in self.commands:
            commandInfo[0].errback(error)
        self.commands = []
        
    def runPOPCommand(self, command, *args, **kwargs):
        print 'running POP command:', command
        d = defer.Deferred()
        self.commands.append([d, command, args, kwargs])
        print "added to queue."
        print self.commands
        if not self.client:
            #FIXME: refresh list
            reactor.connectTCP(self.url.server, int(self.url.port), self)
        elif not self.busy: 
            self.runNextCommand()
        return d
            
    def runNextCommand(self):
        if self.commands:
            d, command, args, kwargs = self.commands[0]
            func = getattr(self.client, command)
            self.busy = 1
            apply(func, args, kwargs)
        else:
            self.busy = 0
        
    def commandCompleted(self, command, results):
        print "Command completed", command, results
        if self.commands[0][1] == command:
            d, command, args, kwargs = self.commands.pop(0)
            d.callback(results)
            self.runNextCommand()
        #else:
            #raise "Got unexpected response!"
        reactor.callLater(30, self.disconnectIfInactive)
        
    def disconnectIfInactive(self):
        if self.client and not self.commands:
            print "Disconnecting!"
            self.client.quit()
        
    def listMessages(self, path=None):
        d = self.runPOPCommand('uidl')
        d.addCallback(self.gotUIDL)
        return d
        
    def gotUIDL(self, keyList):
        self.messageMap = {}
        for messageNo, key in keyList:
            self.messageMap[key] = messageNo
        # remove messages from cache if they're no longer
        # on the server.
        for cachedKey in self.cache.keys():
            if not self.messageMap.has_key(cachedKey.replace('msg_', '', 1)):
                del(self.cache[cachedKey])
        return self.messageMap.keys()

    def getMessage(self, messageKey, path=None):
        if self.cache.has_key('msg_' + messageKey):
            return defer.succeed(
                message.Message(self.cache['msg_' + messageKey])
                )
        else:    
            d = self.runPOPCommand('retr', self.messageMap[messageKey])
            d.addCallback(self._gotMessage, messageKey)
            return d
        
    def _gotMessage(self, stream, key):
        msgData = stream.read()
        self.cache['msg_' + key] = msgData
        return message.Message(msgData)
        
    def deleteMessage(self, messageKey, path=None):
        print "POP: deleting."
        return self.runPOPCommand('dele', self.messageMap[messageKey])

    def buildProtocol(self, addr):
        if self.client:
            raise "Client already exists!!!"
        client = BasicPOP3()
        client.factory = self
        return client
            
    def clientConnectionFailed(self, connection, reason):
        self.connectionFailed(reason)
        
class BasicPOP3(POP3Client):
    "logs in and then waits for input."

    def handle_WELCOME(self, line):
        # start login to server
        self.user(self.factory.url.username)
    
    def handle_USER(self, line):
        if line.upper().startswith('-ERR'):
            self.factory.connectionFailed("No Such User")
            self.quit()
            return
        self.pass_(self.factory.url.password)

    def handle_PASS(self, line):
        if line.upper().startswith('-ERR'):
            # FIXME: use AuthenticationError here (implement them in yarn first)
            self.factory.connectionFailed("Bad Password")
            self.quit()
        else:
            self.factory.client = self
            self.factory.runNextCommand()
        
    def handle_LIST(self, stream):
        lines = [line.strip().split() for line in stream.read().strip().split('\n')]
        stream.close()
        self.factory.commandCompleted('list', lines)
        
    def handle_UIDL(self, stream):
        lines = [line.strip().split() for line in stream.read().strip().split('\n')]
        if not max(lines): lines = []
        stream.close()
        self.factory.commandCompleted('uidl', lines)

    def handle_RETR(self, stream):
        self.factory.commandCompleted('retr', stream)
            
    def handle_DELE(self, response):
        self.factory.commandCompleted('dele', response)

    def connectionLost(self, reason):
        self.factory.client = None
        self.factory.messageMap = {}
