# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re
import time
import formats
from url import Url
from twisted.internet import reactor, defer

def fixPath(pathStr):
    "clean up a path String, removing leading/trailing/suplicate slashes"
    while pathStr.count('//'):
        pathStr = pathStr.replace('//', '/') 
    while pathStr.startswith('/'):
        pathStr = pathStr[1:]
    while pathStr.endswith('/'):
        pathStr = pathStr[:-1]
    return pathStr

def joinPath(parent, child):
    "join a parent path to a child path"
    return fixPath("/".join([parent, child]))

class Proxy(object):
    """
    A proxy provides a common interface to a local or remote message
    store.  A proxy can support reading, writing, editing and/or deleting
    messages.  It also may provide a way to open proxies to child paths.    
    """
    protocols = [] # list of protocols this Proxy can handle
    ports = {}     # dict of protocol:port - default port info

    def __init__(self, url_, connector, credentials=None):
        self.url = Url(url_)
        self.connector = connector
        self.credentials = credentials
        self.cache = self.connector.openCache(self.url)
        self.currentViews = []
        
    def connect(self):
        # connect to the URL, return a deferred.
        raise NotImplementedError

    def __del__(self):
        if self.cache:
            del(self.cache) # this should close any open files, etc....

    def getName(self, path):
        return path.split('/')[-1]

    def getFeatures(self, path):
        return []
                    
    def listMessages(self, path):
        """
        return a deferred list of messagekeys.  Each message key
        must be unique to a single message.  Keys must remain valid
        between sessions, unless the message is deleted.
        """
        return defer.succeed([])

    def getLinks(self, path):
        return defer.succeed([])
        
    def getMessage(self, messageKey):
        "return a deferred message identified by messageKey"
        raise NotImplementedError
        
    def editMessage(self, messageKey, newMessage):
        """
        replace the message identified by messageKey with the new Message.
        the messageKey should still be valid for the edited message.
        """
        raise NotImplementedError
        
    def deleteMessage(self, messageKey):
        "delete the specified message"
        raise NotImplementedError
        
    def registerView(self, path, view):
        if (path, view) not in self.currentViews:
            self.currentViews.append((path, view))
            
    def unregisterView(self, path, view):
        while (path, view) in self.currentViews:
            self.currentViews.remove((path, view))
            
    def handleEvent(self, path, event, messageKey):
        for p, view in self.currentViews:
            if p == path:
                view.handleEvent(event, messageKey)
                
    def listPaths(self, path):
        "return a deferred list of child paths"
        return defer.succeed([])

    def createFolder(self, parentPath, folderName):
        "create a new folder under parentPath."
        raise NotImplemetedError, "This proxy does not support createPath."

    def getUrl(self, path):
        "return the fully-qualifed url of path"
        if path == "":
            return self.url
        else:
            raise NotImplementedError, "can't getUrl for path %s." % path
        
    def close(self):
        "shut down proxy, perform any clean-up actions"
        return defer.succeed(1)
         
    def createLink(self, path, link):
        raise NotImplementedError

    def deleteLink(self, path, link):
        raise NotImplementedError

    def editLink(self, path, oldLink, newLink):
        raise NotImplementedError


class SimpleProxy(Proxy):
    """
    A simple implementation of a Proxy.  This class has a single method
    meant to be overridden, getFile(), which should callback with a tuple
    of stream, metadata.
    """
    refreshrate = 60*60 # default: 1 hour.  Not suitable for all purposes.
    
    def __init__(self, url, connector, credentials=None):
        Proxy.__init__(self, url, connector, credentials)
        self.messages = {}
        self.timestamp = None
        #self.properties = {'name':'', 'description': ''}
        
    def connect(self):
        """
        Override this in subclass to fill up self.messages,
        self.properties, self.paths.
        """
        d = self.getFile()
        d.addCallback(self._gotFile)
        return d
        
    def _gotFile(self, results):
        self.timestamp = time.time()
        file, metadata = results
        self.parser = formats.parseFile(file, metadata)
        if not self.parser: raise "UnparseableFileError", "Unknown format"
        
    def callAfterLoading(self, command, *args, **kwargs):
        path = args[0]
        if time.time() - self.timestamp > self.refreshrate:
            d = defer.Deferred()
            self.connect().addCallback(
                self.loaded, d, command, args, kwargs).addErrback(d.errback)
            return d
        else:
            method = getattr(self.parser, command)
            return defer.maybeDeferred(method, *args, **kwargs)
            
    def loaded(self, returnCode, d, command, args, kwargs):
        print "PROXY: reloaded."
        method = getattr(self.parser, command)
        defer.execute(method, *args, **kwargs).chainDeferred(d)
        
    def listMessages(self, path):
        return self.callAfterLoading("listMessages", path)
    def getMessage(self, path, messageKey):
        return self.callAfterLoading("getMessage", path, messageKey)
    def getHeaders(self, path, messageKey):
        return self.callAfterLoading("getHeaders", path, messageKey)
    def getProperties(self, path):
        return self.callAfterLoading("getProperties", path)
    def createLink(self, path, link):
        return self.callAfterLoading("createLink", path, link) 
    def getLinks(self, path):
        return self.callAfterLoading("getLinks", path)
    def listPaths(self, path):
        return self.callAfterLoading("listPaths", path)
    def deleteLink(self, path, link):
        return self.callAfterLoading("deleteLink", path, link)
    def editLink(self, path, oldLink, newLink):
        return self.callAfterLoading("editLink", path, oldLink, newLink)
    def deleteFolder(self, path, folder):
        return self.callAfterLoading("deleteFolder", path, folder)
    

    def getUrl(self, path):
        if not path:
            return self.url
        else:
            return self.parser.getUrl(path)

    def getName(self, path):
        return self.parser.getName(path)

    # get features from parser.
    def getFeatures(self, path):
        return self.parser.getFeatures(path)
