# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
classes to make working with remote message sources easier.
"""

import os.path
import sys
import time
from yarn import proxy, auth, events, protocols
from url import Url, Link
from twisted.internet import defer, reactor
from twisted.persisted import dirdbm
from xml.sax import make_parser
from xml.sax.handler import ContentHandler
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl
from cStringIO import StringIO
import types
import os
import md5


class Folder(object):
    def __init__(self, dataStore=None, keychain=None):
        self.dataStore = dataStore or MetadataStore() # null data store
        if keychain is None: keychain = auth.Keychain()
        self.keychain = keychain
        self.subFolders = {}
        self.proxy = None
        self.path = ""
        self.eventHandlers = {}
        self.links = {} # dict of foldername: link

    def connect(self, url, connector=None, loadFolders=True):
        """
        Connect this Folder (via a proxy) to the provided Url.  This method
        must be called before the Folder can be used to read/write/list
        messages.
        """
        self.connector = connector or Connector()
        self.url = url
        self._credentials = self.keychain.get(
            str(url)) or auth.AnonymousCredentials()
        
        d = defer.Deferred()
        self.connector.openProxy(
            self.url, self._credentials).addCallback(
            self._gotProxy, d, loadFolders).addErrback(
            self._openProxyError, d)
        return d
                   
    connected = property(lambda self: self.proxy and True or False)

    def _openProxyError(self, error, connectD):
        connectD.errback(error)
        return 1
    
    def _gotProxy(self, proxy, connectD, loadFolders):
        self.connectProxy(proxy, "")
        if loadFolders:
            self.refresh().addCallback(
                lambda result: connectD.callback(result)).addErrback(
                lambda err: connectD.errback(err))
        else:
            connectD.callback(1)

    def connectProxy(self, proxy, path):
        self.proxy = proxy
        self.path = path
        self.url = self.proxy.getUrl(path)
        self.summary = FolderSummary(self)
        self.proxy.registerView(self.path, self)
        # make sure metadata contains this url
        if not self.dataStore.urls.has_key(self.url):
            self.dataStore.urls[self.url] = {}
            self.dataStore.messages[self.url] = {}
        self.metadata = self.dataStore.urls[self.url]
        self.messageMetadata = self.dataStore.messages[self.url]

    features = property(lambda self: self.proxy.getFeatures(self.path))
        
    def _getName(self):
        return self.metadata.get('name','')
    def _setName(self, name):
        self.metadata['name'] = name
    name = property(_getName, _setName)

    def _proxyMethod(self, method, *args, **kwargs):
        "run a method on self.proxy and return the results."
        if not self.proxy: raise Exception, "Not connected."
        method = getattr(self.proxy, method)
        return method(self.path, *args, **kwargs)

    def deleteMessage(self, messageKey):
        self._proxyMethod("deleteMessage", messageKey)
        
    def editMessage(self, messageKey, newMessage):
        return self._proxyMethod("editMessage", messageKey, newMessage)
        
    def postMessage(self, message):
        return self._proxyMethod("postMessage", message)

    def listMessages(self):
        d = self._proxyMethod("listMessages")
        d.addCallback(self._recordMessageKeys)
        return d
        
    def _recordMessageKeys(self, messageKeys):
        "track the keys we've seen"
        if not messageKeys: return messageKeys
        changes = 0
        for key in messageKeys:
            if not self.messageMetadata.has_key(key):
                self.messageMetadata[key] = {}
                changes += 1                
        # remove metadata for messages that no longer exist
        #for mUrl, mKey in self.messageProperties.keys():
        #    if mUrl == url and mKey not in messageKeys:
        #        del(self.messageProperties[(mUrl, mKey)])
        #        changes = 1
        if changes: self.dataStore.saveSoon()
        return messageKeys

    def getMessage(self, messageKey):
        d = self._proxyMethod("getMessage", messageKey)
        d.addCallback(self._inspectMessage,  messageKey)
        return d

    def _inspectMessage(self, message, key):
        "record information about this message for future use."
        # save the original timestamp, if we haven't seen this message before
        # FIXME: use Received: header and current time.
        ts = self.messageMetadata[key].setdefault('timestamp', message['date'])
        message['X-Yarn-Timestamp'] = ts
        message['X-Yarn-Source-Url'] = str(self.url)
        return message

    def getMessages(self):
        d = defer.Deferred()
        self.listMessages().addCallback(
            self._gotListToGet, d).addErrback(
            d.errback)
        return d
        
    def _gotListToGet(self, keys, finished):
        getters = []
        for k in keys:
            getters.append(self.getMessage(k))
        defer.DeferredList(getters).addCallback(
            self._cleanupGotMessages).addCallback(
            finished.callback).addErrback(
            finished.errback)
        
    def _cleanupGotMessages(self, messageValues):
        return [m for success, m in messageValues]

    def createLink(self, link):
        d = defer.Deferred()
        self._proxyMethod("createLink", link).addCallback(
            self._createdLinkOrFolder, d).addErrback(d.errback)
        return d

    def createFolder(self, folderName):
        d = defer.Deferred()
        self._proxyMethod("createFolder", folderName).addCallback(
            self._createdLinkOrFolder, d).addErrback(d.errback)
        return d

    def _createdLinkOrFolder(self, newLinkOrFolder, finished):
        if isinstance(newLinkOrFolder, Link):
            finished.addCallback(lambda _: newLinkOrFolder.title)
            self._openLinkedFolder(newLinkOrFolder).chainDeferred(finished)
        else: # it's a path
            self._loadFolder(newLinkOrFolder)
            finished.callback(newLinkOrFolder)

    def editLink(self, oldLink, newLink):
        finished = defer.Deferred()
        self._proxyMethod("editLink", oldLink, newLink).addCallback(
            self._editedLink, oldLink, finished).addErrback(finished.errback)
        return finished

    def _editedLink(self, newLink, oldLink, finished):
        # remove and re-create the folder object, in case the name
        # has changed.
        if self.subFolders.has_key(oldLink.title):
            del(self.subFolders[oldLink.title])
        del(self.links[oldLink.title])
        self._createdLinkOrFolder(newLink, finished)

    def deleteFolder(self, folder):
        "delete the specified folder."
        # FIXME: this should work on links, too
        if self.isLink(folder):
            d = self._proxyMethod("deleteLink", self.links[folder])
        else:
            d = self._proxyMethod("deleteFolder", folder)
        d.addCallback(self._deletedLinkOrFolder, folder)
        return d

    def _deletedLinkOrFolder(self, result, folder):
        # the folder no longer exists, so remove the local reference
        if self.subFolders.has_key(folder): del(self.subFolders[folder])
        if self.links.has_key(folder): del(self.links[folder])

    def getAlternateLinks(self):
        d = defer.Deferred()
        self._proxyMethod("getLinks").addCallback(
            self._lookupAlternateLinks, d).addErrback(d.errback)
        return d

    def _lookupAlternateLinks(self, links, finished):
        alternates = []
        recursors = []
        for link in links:        
            if link.properties.get('recurseInto'):
                f = Folder() # not passing self.dataStore, don't save metadata
                f.keychain = self.keychain
                d = defer.Deferred()
                f.connect(link.url, self.connector).addCallback(
                    self._connectedToAlternateUrl, f, d).addErrback(d.errback)
                recursors.append(d)
            elif link.rel == 'alternate':
                alternates.append(link)
        defer.DeferredList(recursors).addCallback(
            self._gotAlternateLinks, alternates, finished).addErrback(
            finished.errback)
        # add errbacks to each deferred to clean up errors
        for d in recursors: d.addErrback(lambda err: None)

    def _connectedToAlternateUrl(self, result, folder, finished):
        folder.getAlternateLinks().chainDeferred(finished)

    def _gotAlternateLinks(self, results, links, finished):
        for succeeded, result in results:
            if succeeded:
                links += result
            else:
                pass # ignore errors
        finished.callback(links)

    def refresh(self):
        # this will refresh our list of current messages
        getInfo = defer.DeferredList([
            self._refreshFolders(), # gives a base list of paths to start from
            self.summary.refresh(),
            self.refreshMetadata()
            ])
        getInfo.addCallback(self._refreshed)
        return getInfo

    def _refreshed(self, val):
        pass

    def _refreshFolders(self):
        finished = defer.Deferred()
        # clear out existing paths
        for subPath in self.subFolders.keys():
            del(self.subFolders[subPath])
        self.links = {}
        pathLister = self.proxy.listPaths(self.path)
        linkGetter = self.proxy.getLinks(self.path)
        defer.DeferredList([pathLister, linkGetter],
                           fireOnOneErrback=1).addCallback(
            self._gotPathsAndLinks, finished).addErrback(finished.errback)
        return finished

    def _gotPathsAndLinks(self, results, finished):
        paths, links = results[0][1], results[1][1]

        # set up sub-Folder objects
        for path in paths:
            self._loadFolder(path)

        # refresh all sub-Folders, callback when done.
        refreshers = []
        loaders = []
        for f in self.subFolders.values():
            refreshers.append(f.refresh())
        for link in links:
            if link.rel == 'child':
                loaders.append(self._openLinkedFolder(link))
        refreshing = defer.DeferredList(refreshers+loaders)
        refreshing.chainDeferred(finished) # send finished callbacks, errbacks
        # catch residual errors
        for d in refreshers+loaders: d.addErrback(self._printError)
        return refreshing

    def _printError(self, err):
        print "FOLDER: error loading sub-folder:", err.getErrorMessage()
        #import traceback; traceback.print_exc()

    def refreshMetadata(self):
        self.metadata['name'] = self.proxy.getName(self.path)
        self.dataStore.saveSoon()
        return defer.succeed(None)

    def _loadFolder(self, path):
        "add the folder at self.path/path to self.subFolders"
        fullPath = proxy.joinPath(self.path, path)
        f = Folder(self.dataStore, self.keychain)
        f.connectProxy(self.proxy, fullPath)
        self.subFolders[path] = f

    def _openLinkedFolder(self, link):
        """
        open a Folder to the given url, and file it in self.subFolders
        under the given name.
        """
        finished = defer.Deferred()
        f  = Folder(self.dataStore, self.keychain)
        # track the link behind the folder
        self.links[link.title] = link
        f.connect(
            link.url, self.proxy.connector, loadFolders=False).addCallback(
            self._gotLinkedFolder, f, link, finished).addErrback(
            finished.errback)
        return finished

    def _gotLinkedFolder(self, result, folder, link, finished):
        self.subFolders[link.title] = folder
        # refresh the folder now that we're connected.
        self.subFolders[link.title].refresh().chainDeferred(finished)

    def isLink(self, folderName):
        return self.links.has_key(folderName)

    def setEventHandler(self, event, function):
        self.eventHandlers[event] = function

    def clearEventHandler(self, event):
        if self.eventHandlers.has_key(event): del(self.eventHandlers[event])

    def handleEvent(self, event, messageKey):
        """
        if an event handler is registered for this event, run
        the handler function with the supplied arguments.
        """
        if event == events.MESSAGE_POSTED:
            # see if we already know of this message
            if self.messageMetadata.has_key(messageKey):
                return
            else:
                # don't care when this completes, just start the process now
                self.summary.refresh()
        elif event == events.MESSAGE_DELETED:
            # FIXME: be smarter
            self.summary.refresh()
        handler = self.eventHandlers.get(event)
        if handler: handler(messageKey)

    credentialsClass = property(lambda self: self.proxy.credentialsClass)

    def save(self): self.dataStore.save()
    def saveSoon(self): self.dataStore.saveSoon()

class FolderMetadataSaxHandler(ContentHandler):
    
    def __init__(self, data):
        self.data = data
        self.stack = []
        self.currentUrl = ""
        self.currentNode = None

    def startElement(self, name, attrs):
        self.stack.append(name)
        if name == 'folder':
            self.currentUrl = attrs['url']
            self.currentNode = self.data.urls[self.currentUrl] = {}
        elif name == 'messages':
            self.data.messages[self.currentUrl] = {}
        elif name == 'message':
            self.currentNode = self.data.messages[
                self.currentUrl][attrs['key']] = {}
        elif name == 'property':
            val = attrs.get('value', '')
            type = attrs.get('type', 'string')
            converter = self.data.converters[type]
            self.currentNode[attrs['name']] = converter(val)

    def endElement(self, name):
        self.stack.pop()

class StringDict(dict):
    """dict that converts all keys into strings."""
    def __setitem__(self, key, val):
        return dict.__setitem__(self, str(key), val)

    def __getitem__(self, key):
        return dict.__getitem__(self, str(key))

    def has_key(self, key):
        return dict.has_key(self, str(key))

class MetadataStore(object):
    """
    Metadata storage for Folders.
    """

    types = {
        types.UnicodeType: 'string',
        types.IntType: 'int',
        }
    converters = {
        'string': unicode,
        'int': int,
        }

    def __init__(self, filename=None):
        self.saving = None
        self._filename = filename
        self.urls = StringDict()
        self.messages = StringDict()
        if filename and os.path.exists(filename):
            self.load(filename)

    def load(self, filename):
        parser = make_parser()
        handler = FolderMetadataSaxHandler(self)
        parser.setContentHandler(handler)
        parser.parse(file(filename))
        
    def saveSoon(self):
        "Register the need for the folder to be saved in the near future."
        if self._filename and not self.saving: 
            self.saving = reactor.callLater(1, self.save)

    def _dictToXml(self, dict, xml):
        for key, val in dict.items():
            # note the type of the object, defaulting to string
            valType = self.types.get(type(val), 'string')
            if valType != 'string': # str-ify non-string values
                val = str(val)
            # convert everything to unicode.
            if type(val) != types.UnicodeType:
                val = unicode(val)
            if type(key) != types.UnicodeType:
                key = unicode(key)
            xml.startElement('property', {'name': key.encode('utf-8'),
                                          'type': valType,
                                          'value' :val.encode('utf-8')
                                          ,})
            xml.endElement('property')        

    def save(self):
        if not self._filename: return
        encoding = 'utf-8'
        f = StringIO()
        xml = XMLGenerator(f, encoding)
        xml.startElement('view', {})
        for url, urlData in self.urls.items():
            xml.startElement('folder', {'url':url})
            self._dictToXml(urlData, xml)
            xml.startElement('messages', {})
            for key, messageData in self.messages[url].items():
                xml.startElement('message', {'key':key})
                self._dictToXml(messageData, xml)
                xml.endElement('message')
            xml.endElement('messages')
            xml.endElement('folder')
        xml.endElement('view')
        f.seek(0)
        output = file(self._filename, 'w+b')
        output.writelines(['<?xml version="1.0" encoding="%s"?>' % encoding,
                           f.read(),
                           ''])
        output.close()
        self.saving = None

class FolderSummary(object):
    def __init__(self, folder):
        self._folder = folder
        self.messageKeys = []

    def _getMessageCount(self):
        return len(self.messageKeys)
    count = property(_getMessageCount)

    def _getUnseenCount(self):
        count = 0
        for key in self.messageKeys:
            if not self._folder.messageMetadata[key].get('seen'):
                count += 1
        return count
    unseenCount = property(_getUnseenCount)
    
    def refresh(self):
        return self._folder.listMessages().addCallback(
            self._gotKeys)

    def _gotKeys(self, keys):
        self.messageKeys = keys

class Connector:
    def __init__(self, cacheDir=None):	    
        self.preferences = {}
        if cacheDir:
            self.cacheDir = os.path.abspath(cacheDir)
        else:
            # create a cache dir in the temp directory
            # FIXME: think about horrible security implication of doing this
            import tempfile
            self.cacheDir = os.path.join(tempfile.gettempdir(), "%s-%s-cache" % (os.getpid(), time.time()))
        if not os.path.isdir(self.cacheDir): os.mkdir(self.cacheDir)
        self.offline = 0
                
        self._proxies = {}
        self._openProxyCallbacks = []
        self._loadingProxies = []
    
    def openCache(self, url, cacheClass=dirdbm.Shelf):
        cacheDir = os.path.join(self.cacheDir, (md5.md5(str(url)).hexdigest()))
        if not os.path.exists(cacheDir): os.mkdir(cacheDir)
        return cacheClass(cacheDir)

    def openProxy(self, urlStr, credentials=None):
        #print "CONNECTOR: openProxy", urlStr, credentials
        if not credentials: credentials = auth.AnonymousCredentials()
        urlStr = str(urlStr) # in case it's a URL object
        resourceKey = (urlStr, credentials.getDigest())
        if self._proxies.has_key(resourceKey):
            return defer.succeed(self._proxies[resourceKey])
        else:
            d = defer.Deferred()
            # if we're not already loading the url, load it.
            if resourceKey not in self._loadingProxies:
                u = Url(urlStr)
                self._loadingProxies.append(resourceKey)
                self._openProxyCallbacks.append((resourceKey, d))
                protocols.openProxy(self, u, credentials).addCallback(
                    self.gotProxyFromUrl, resourceKey).addErrback(
                    self.gotProxyFromUrl, resourceKey, False)
            else:
                self._openProxyCallbacks.append((urlStr, d))
            return d

    def gotProxyFromUrl(self, proxy, resourceKey, succeeded=True):
        if succeeded:
            self._proxies[resourceKey] = proxy
        self._loadingProxies.remove(resourceKey)
        for key, d in self._openProxyCallbacks:
            if key == resourceKey:
                if succeeded:
                    d.callback(proxy)
                else:
                    d.errback(proxy) # 'proxy' actually contains an error
        for callback in self._openProxyCallbacks:
            if callback[0] == resourceKey:
                self._openProxyCallbacks.remove(callback)
    
    def close(self):
        return defer.DeferredList([p.close() for p in self._proxies.values()])
        

