# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"Tests for yarn.message.Message objects."

from twisted.trial import unittest
from yarn import message
import os

class NewMessageTester(unittest.TestCase):
    strings = [
        'hello world',
        u'hello world',
        u'\u00bfHabla espa\u00f1ol?',
        u'\u65e5\u672c\u8a9e',
        ]
    encodings = ['utf-8', 'utf-7', 'latin-1', None]
    
    def testTitle(self):
        for string in self.strings:
            m = message.Message()
            m['Subject'] = string
            self.assertEquals(m['Subject'], string)

    def testText(self):
        for string in self.strings:
            m = message.Message('text/plain')
            m.contents = string
            self.assertEquals(m.contents, string)
        
    def testHtml(self):
        for string in self.strings:
            m = message.Message('text/html')
            m.contents = string
            self.assertEquals(m.contents, string)

    def testChangeContents(self):
        for string1 in self.strings:
            for string2 in self.strings:
                m = message.Message('text/plain')
                m.contents = string1
                self.assertEquals(m.contents, string1)
                m.mimetype = 'text/html'
                self.assertEquals(m.contents, string1)
                m.contents = string2
                self.assertEquals(m.contents, string2)

    def testGenerateMissingParts(self):
        for string in self.strings:
            m = message.Message(mimetype='text/html')
            m.contents = string
            m['subject'] = string
            m['x-yarn-link'] = 'http://www.site.com/page'
            copy = m.getCopy(generateMissingParts=True)
            self.assertEquals(len(copy.alternatives), 2)
            self.assertEquals([a.mimetype for a in copy.alternatives],
                                  ['text/plain', 'text/html'])
            self.assertEquals(
                copy.alternatives[0].contents.count(m['x-yarn-link']), 1)
            self.assertEquals(
                copy.alternatives[1].contents.count(m['x-yarn-link']), 2)

    def testDefaultContents(self):
        m = message.Message()
        m.contents = 'hello world!'
        alt = message.MessageContents(mimetype='text/html')
        alt.mimetype = 'text/html'
        alt.contents = '<b>Hello</b> World'
        m.alternatives.append(alt)
        self.assertEquals(len(m.alternatives), 2)
        self.assertEquals(m.contents, '<b>Hello</b> World')
        self.assertEquals(m.mimetype, 'text/html')



class EmailTester(unittest.TestCase):
    # list of filename, types, attachmentCount
    testFiles = [
        ('pgp-signed.eml', ['text/plain'], 1),
        ('multipart-html.eml', ['text/plain', 'text/html'], 0),
        ('html-with-attachment.eml', ['text/html'], 1),
        ('multipart-with-attachment.eml', ['text/plain', 'text/html'], 1),
        ]

    def testParseEmails(self):
        emailDir = os.path.join(os.path.split(__file__)[0], 'files', 'email')
        for emailFile, types, attachmentCount in self.testFiles:
            emailFilePath = os.path.join(emailDir, emailFile)
            msg = message.Message(file(emailFilePath))
            # make sure we can read message contents.
            self.assert_(len(msg.contents) > 0)
            # check alternatives
            self.assertEquals([a.mimetype for a in msg.alternatives],
                              types)
            # try editing the message (should wipe out alternatives)
            msg.contents = 'hello world'
            msg.mimetype = 'text/plain'
            self.assertEquals(len(msg.alternatives), 1)
            self.assertEquals(msg.contents, 'hello world')
            self.assertEquals(msg.alternatives[0].contents, 'hello world')
            self.assertEquals(msg.mimetype, 'text/plain')


class AddressTester(unittest.TestCase):
    testNames = [
        u'\u00bfHabla espa\u00f1ol?',
        u'\u65e5\u672c\u8a9e',
        'Bob'
        ]

    def testEncode(self):
        for s in self.testNames:
            m = message.Message()
            addr = message.NameAndAddress()
            addr.name = s
            addr.address = 'bob@somewhere.bob'
            m['From'] = addr
            self.assertEquals(message.NameAndAddress(m['From']).address,
                              'bob@somewhere.bob')
            self.assertEquals(message.NameAndAddress(m['From']).name, s)
            
