# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.trial import unittest
from twisted.trial.util import deferredResult
from twisted.web import xmlrpc, server
from twisted.internet import reactor
from twisted.python import failure
from yarn import url, auth, remote, errors, message
import viewtests

blogUser = 'user'
blogPass = 'pass'
blogName = 'My Cool Blog'
blogData = [
    ("Entry One", "I like cheese."),
    ("Entry Two", """
                  My favorite things:
                  <ul>
                  <li>Cheese</li>
                  <li>Sausage</li>
                  <li>Crackers</li>
                  </ul>""")
    ]

class RPCResource(xmlrpc.XMLRPC):

    def __init__(self, blogData, *args):
        xmlrpc.XMLRPC.__init__(self, *args)
        self.putSubHandler("blogger", BloggerResource(blogData))
        self.putSubHandler('metaWeblog', MetaWeblogResource(blogData))

    def render(self, *args, **kwargs):
        request = args[0]
        try:
            return xmlrpc.XMLRPC.render(self, *args, **kwargs)
        except Exception, e:
            #request.content.seek(0,0)
            #file('/tmp/request.xml', 'w+b').write(request.content.read())
            print "method failed:", e
            raise e
        
class BloggerResource(xmlrpc.XMLRPC):

    def __init__(self, data):
        self.blogData = data
        xmlrpc.XMLRPC.__init__(self)

    def xmlrpc_getUsersBlogs(self, appkey, username, password):
        # only check the password here, for testing purposes

        if username != blogUser or password != blogPass:
            return xmlrpc.Fault(0, "Invalid username or bad password")
        else:
            return [{'blogid': 1, 'blogName':blogName}]

    def xmlrpc_getUserInfo(self, appkey, username, password):
        return {'userid': 1, 'firstname':'Joe', 'lastname': 'Blogger'}

class MetaWeblogResource(xmlrpc.XMLRPC):

    def __init__(self, data):
        self.blogData = data
        xmlrpc.XMLRPC.__init__(self)

    def xmlrpc_getRecentPosts(self, blogid, username, password, count):
        entries = []
        # ignoring the count for the moment.
        for id in range(len(self.blogData)):
            entries.append(self.xmlrpc_getPost(id, username, password))
        return entries
                       
    def xmlrpc_getPost(self, postid, username, password):
        data = self.blogData[postid]
        link = 'http://www.site.com/' + data[0].lower().replace(' ', '_')
        struct = {'dateCreated': xmlrpc.DateTime(),
                  'userid': 1,
                  'postid': postid,
                  'title': data[0],
                  'description': data[1],
                  'link': link,
                  }
        return struct

    def xmlrpc_newPost(self, blogid, username, password, post, publish):
        try:
            postData = (post['title'], post['description'])
            self.blogData.append(postData)
            return len(self.blogData) - 1
        except Exception, e:
            print "SERVER ERROR!!!", e
            import sys
            print sys.exc_info()
            print dir(e)
            raise e

class MetaWeblogTester(unittest.TestCase, viewtests.FolderTesterMixin):
    PORT = 9982
    
    def setUp(self):
        website = server.Site(RPCResource(blogData[:]))
        self._port = reactor.listenTCP(self.PORT, website)
        self.url = 'metaweblog:http://localhost:%i/' % (self.PORT)
        for x in range(2): reactor.iterate(0.1)

    def tearDown(self):
        self._port.stopListening()
        for x in range(2): reactor.iterate(0.1);
        del(self._port)

    def getUrl(self): return self.url

    def getCredentials(self):
        return auth.UserPassCredentials(blogUser, blogPass)

    def testTester(self):
        "Test that the text RPC server is working..."
        rpcUrl = url.Url(self.url).path
        s = xmlrpc.Proxy(rpcUrl)
        d = s.callRemote("metaWeblog.getRecentPosts",
                         1, blogUser, blogPass, 10)
        posts = deferredResult(d)
        #self.assertEquals(len(posts), 2)
       
    def testFetch(self):
        "test fetching messages"
        folder = self.getConnectedFolder()
        path = folder.subFolders.keys()[0]
        self.assertEquals(path, blogName)
        blogFolder = folder.subFolders[path]

        keys = deferredResult(blogFolder.listMessages())
        self.assertEquals(keys, [0,1])

        for key in keys:
            message = deferredResult(blogFolder.getMessage(keys[key]))
            self.assertEquals(message['Subject'], blogData[key][0])
            self.assertEquals(message.contents, blogData[key][1])

if __name__=="__main__":
    website = server.Site(RPCResource())
    print "RUNNING AT http://localhost:8080/"
    reactor.listenTCP(8080, website)
    reactor.run()
