# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.trial import unittest
from yarn.url import Url
from yarn import protocols # FIXME: this shouldn't be needed!

class UrlTester(unittest.TestCase):
    
    def _testUrl(self, url, protocol, server, path,
                 args='', user='', password='', fragment=''):
        u = Url(url)
        self.assertEquals(u.protocol, protocol)
        self.assertEquals(u.server, server)
        self.assertEquals(u.path, path)
        self.assertEquals(u.args, args)
        self.assertEquals(u.fragment, fragment)

    def testEquals(self):
        self.assertEquals('http://www.site.com',
                          Url('http://www.site.com'))
        self.assertEquals(['http://www.site1.com',
                           'http://www.site2.com'],
                          [Url('http://www.site1.com'),
                           Url('http://www.site2.com')]
                          )

    def testHTTP(self):
        for u in ['http://www.slashdot.org/page',
                  'www.slashdot.org/page']:
            self._testUrl(u, 'http', 'www.slashdot.org', '/page')
            
        self._testUrl('127.0.0.1', 'http', '127.0.0.1', '')

    def testHTTPArgs(self):
        self._testUrl('http://www.slashdot.org/page?arg1=val',
                        'http', 'www.slashdot.org', '/page', 'arg1=val')
                 
    def testMailto(self):
        self._testUrl('mailto:billg@hotmail.com', 'mailto',
                      'hotmail.com', '', user='billg')

    def testFile(self):
        u = Url(__file__)
        self.assertEquals(u.protocol, 'file')

    def testFile2(self):
        self._testUrl('/my/file', 'file', '', '/my/file')
        self._testUrl(r'c:\program files\test?stuff\test.exe',
                      'file', '', r'c:\program files\test?stuff\test.exe')

    def testFile3(self):
        self.assertEquals(Url('/my/file'), 'file:///my/file')
        self.assertEquals(Url(r'c:\my dir\file'), r'file://c:\my dir\file')

    def testQuoting(self):
        self.assertEquals(Url('http://site.com/path with spaces'),
                          Url('http://site.com/path%20with%20spaces'))
        self.assertEquals(Url('http://site.com/path+with+spaces'),
                          Url('http://site.com/path%20with%20spaces'))
        self._testUrl('http://site.com/path%20with%20spaces',
                      'http', 'site.com', '/path with spaces')

    def testCompositeUrl(self):
        u = Url('atom:https://www.site.com/atom')
        self.assertEquals(u, 'atom:https%3A//www.site.com/atom')
        self.assertEquals(u.protocol, 'atom')
        u2 = Url(u.path)
        self.assertEquals(u2.protocol, 'https')
        self.assertEquals(u2.server, 'www.site.com')

        self._testUrl('atom:https://www.site.com/atom#frag',
                      'atom', '', 'https://www.site.com/atom', fragment='frag')

    def testPorts(self):
        u = Url('http://www.site.com')
        self.assertEquals(u.port, 80)

    def testChanges(self):
        "test that Urls don't change when repeatedly re-encoded"
        tests = [
            'http://www.site.com/~user/',
            'file://c:\\file.txt',
            'http://site.com/page?query=stuff#fragment',
            'composite:http://some.site/page',
            'composite:http://some.site/page#ext',
            ]
        for test in tests:
            original = Url(test)
            for x in range(3):
                test = Url(test)
                self.assertEquals(test, original)


from yarn.auth import UserPassCredentials, AnonymousCredentials
class UrlCredTester(unittest.TestCase):

    def testUser(self):
        "test a url with user name set"
        u = Url('http://user@www.site.com')
        self.assertEquals(u.getCredentials(),
                          UserPassCredentials('user'))

    def testUserPass(self):
        "test a url with username and password"
        u = Url('http://user:pass@www.site.com')
        self.assertEquals(u.getCredentials(),
                          UserPassCredentials('user', 'pass'))

    def testNull(self):
        "test a url with no login information"
        u = Url('http://www.site.com')
        self.assertEquals(u.getCredentials(), AnonymousCredentials())
