# This file is part of yarn, a library for working with messaging systems.
# Copyright (C) 2004 Abe Fettig
#
# yarn is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from yarn import remote, message, events, features, auth, errors
from twisted.trial import util
from twisted.python import failure

class FolderTesterMixin:
    """
    mixin to run a whole suite of tests on a folder, trying out pretty
    much everything the underlying proxy claims to support.  It relies
    on getUrl to return the url to connect to, and getCredentails to
    return credentials for that url.
    """
    messageTests = [
        ('Test', 'Test', None),
        ('Test', 'Test', 'latin-1'),
        ('Test', 'Test', 'utf-8'),
        (u'\u00bfHabla espa\u00f1ol?', u'\u00bfHabla espa\u00f1ol?', None),
        (u'\u00bfHabla espa\u00f1ol?', u'\u00bfHabla espa\u00f1ol?','latin-1'),
        ]

    def getUrl(self): raise NotImplementedError

    def getCredentials(self): return auth.AnonymousCredentials()

    def getFolder(self, withCredentials=True):
        f = remote.Folder()
        if withCredentials:
            keys = auth.Keychain()
            keys[self.getUrl()] = self.getCredentials()
            f.keychain = keys
        return f

    def getConnectedFolder(self, connector=None):
        f = self.getFolder()
        util.deferredResult(f.connect(self.getUrl(), connector))
        return f
    
    def testConnectNoAuth(self):
        """
        try to connect with no credentials. if that returns a valid exception,
        try again, with proper credentials.
        """
        f = self.getFolder(withCredentials=False)
        try:
            util.deferredResult(f.connect(self.getUrl()))
        except failure.Failure, e:
            if e.check(errors.CredentialsRequired):
                f2 = self.getFolder() # this time w/ valid keychain
                util.deferredResult(f2.connect(self.getUrl()))
            else:
                raise e
    
    def testPost(self):
        f = self.getConnectedFolder()

        while not features.POST_MESSAGE in f.features:
            # see if posting to sub folders is supported.
            if f.subFolders:
                f = f.subFolders.values()[0]
            else:
                print "Posting not supported, abandoning test."
                return
        
        keys = util.deferredResult(f.listMessages())
        print "Existing keys:", keys
        for subject, body, charset in self.messageTests:
            m = message.Message('text/html')
            m['Subject'] = subject
            m.contents = body
            # verify that the text got into the message OK.
            self.assertEquals(m['Subject'], subject)
            self.assertEquals(m.contents, body)
            # post the message
            newKey = util.deferredResult(f.postMessage(m))
            print 'posted', newKey
            keys.append(newKey)
            # make sure the keys are listed as expected.
            FolderKeys = util.deferredResult(f.listMessages())
            self.assertEquals(keys, FolderKeys)
            # get a copy of the posted message, make sure it's unchanged.
            m2 = util.deferredResult(f.getMessage(newKey))
            self.assertEquals(m2['Subject'], subject)
            self.assertEquals(m2.contents, body)

    def testMessagePostedEvent(self):
        c = remote.Connector()
        f1 = self.getConnectedFolder(c)
        f2 = self.getConnectedFolder(c)

        if not features.POST_MESSAGE in f1.features:
            print "Posting not supported, abandoning test."
            return

        m = message.Message()
        m['Subject'] = 'Hello World'
        m.contents = 'Test'
        self.assertEquals(f1.summary.count, 0)
        self.assertEquals(f2.summary.count, 0)
        util.deferredResult(f1.postMessage(m))
        # both Folders should be aware of the new message now.
        self.assertEquals(f1.summary.count, 1)
        self.assertEquals(f2.summary.count, 1)
        # now make sure event handlers work
        keys = []
        saveKeyFunc = lambda key: keys.append(key)
        f1.setEventHandler(events.MESSAGE_POSTED, saveKeyFunc)
        newKey = util.deferredResult(f2.postMessage(m))
        self.assertEquals(keys, [newKey])
        self.assertEquals(f1.summary.count, 2)

        
